#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Initial property package for water treatment via a Coagulation-Flocculation process
"""

from pyomo.environ import (
    Constraint,
    Var,
    Param,
    Expression,
    Reals,
    NonNegativeReals,
    Suffix,
    value,
    exp,
    check_optimal_termination,
)

from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
)
from idaes.core.base.components import Component
from idaes.core.base.phases import LiquidPhase
from idaes.core.util.initialization import (
    fix_state_vars,
    revert_state_vars,
    solve_indexed_blocks,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables,
)
from idaes.core.util.exceptions import (
    PropertyPackageError,
    InitializationError,
    ConfigurationError,
)
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
from watertap.core.util.scaling import transform_property_constraints

__author__ = "Austin Ladshaw"

# Set up logger
_log = idaeslog.getLogger(__name__)


# Forward declaration of 'CoagulationParameterData'
@declare_process_block_class("CoagulationParameterBlock")
class CoagulationParameterData(PhysicalParameterBlock):
    CONFIG = PhysicalParameterBlock.CONFIG()

    def build(self):
        """
        Callable method for Block construction.
        """
        super(CoagulationParameterData, self).build()

        self._state_block_class = CoagulationStateBlock

        # phases
        self.Liq = LiquidPhase()

        # components
        self.H2O = Component()
        self.TSS = Component()
        self.TDS = Component()
        self.Sludge = Component()

        #   heat capacity of liquid
        self.cp = Param(initialize=4184, units=pyunits.J / (pyunits.kg * pyunits.K))

        #   reference density of liquid
        self.ref_dens_liq = Param(
            domain=Reals,
            initialize=999.26,
            mutable=True,
            units=pyunits.kg / pyunits.m**3,
            doc="Reference water mass density parameter @ 0 oC and no salts",
        )

        #   change in liquid density with increasing mass fraction of salts/solids
        self.dens_slope = Param(
            domain=Reals,
            initialize=879.04,
            mutable=True,
            units=pyunits.kg / pyunits.m**3,
            doc="Relative increase in liquid density with mass fraction of salts",
        )

        #   adjustment parameters for density change with temperature
        #   Density calculation as a function of temperature and pressure
        #   --------------------------------------------------------------
        #   Engineering Toolbox. Water - Density, Specific Weight, and
        #   Thermal Expansion Coefficients. (2003) https://www.engineeringtoolbox.com/
        #   water-density-specific-weight-d_595.html [Accessed 02-01-2022]
        self.dens_param_A = Param(
            domain=Reals,
            initialize=-2.9335e-6,
            mutable=True,
            units=pyunits.K**-2,
            doc="Density correction parameter A for temperature variation",
        )

        self.dens_param_B = Param(
            domain=Reals,
            initialize=0.001529811,
            mutable=True,
            units=pyunits.K**-1,
            doc="Density correction parameter B for temperature variation",
        )

        self.dens_param_C = Param(
            domain=Reals,
            initialize=0.787973,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Density correction parameter C for temperature variation",
        )

        #   Correction factors for changes in density with changes in pressure
        self.ref_pressure_correction = Param(
            domain=Reals,
            initialize=1.0135,
            mutable=True,
            units=pyunits.dimensionless,
            doc="Density reference correction parameter for changes in pressure",
        )

        self.ref_pressure_slope = Param(
            domain=Reals,
            initialize=4.9582e-10,
            mutable=True,
            units=pyunits.Pa**-1,
            doc="Slope of density change as a function of pressure",
        )

        #   Adjustment for dynamic viscosity as a function of temperature
        #   -------------------------------------------------------------
        #   D.S. Viswananth, G. Natarajan. Data Book on the Viscosity of
        #     Liquids. Hemisphere Publishing Corp. (1989)
        self.mu_A = Param(
            domain=Reals,
            initialize=2.939e-5,
            mutable=True,
            units=pyunits.kg / pyunits.m / pyunits.s,
            doc="Pre-exponential factor for viscosity calculation",
        )

        self.mu_B = Param(
            domain=Reals,
            initialize=507.88,
            mutable=True,
            units=pyunits.K,
            doc="Exponential numerator term for viscosity calculation",
        )

        self.mu_C = Param(
            domain=Reals,
            initialize=149.3,
            mutable=True,
            units=pyunits.K,
            doc="Exponential denominator term for viscosity calculation",
        )

        # ---default scaling---
        self.set_default_scaling("temperature", 1e-2)
        self.set_default_scaling("pressure", 1e-6)

    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {
                "flow_mass_phase_comp": {"method": None},
                "temperature": {"method": None},
                "pressure": {"method": None},
                "mass_frac_phase_comp": {"method": "_mass_frac_phase_comp"},
                "dens_mass_phase": {"method": "_dens_mass_phase"},
                "flow_vol_phase": {"method": "_flow_vol_phase"},
                "conc_mass_phase_comp": {"method": "_conc_mass_phase_comp"},
                "visc_d_phase": {"method": "_visc_d_phase"},
            }
        )

        obj.define_custom_properties(
            {
                "enth_flow": {"method": "_enth_flow"},
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


# this second class contains methods that are applied to state blocks as a whole,
# rather than individual elements of indexed state blocks. Essentially it just includes
# the initialization routine
class _CoagulationStateBlock(StateBlock):
    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:

                         flow_mass_phase_comp : value at which to initialize
                                               phase component flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature

            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states variables are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 release_state method
            outlvl : sets output level of initialization routine (default=idaeslog.NOTSET)
            solver : Solver object to use during initialization if None is provided
                     it will use the default solver for IDAES (default = None)
            optarg : solver options dictionary object (default=None)
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Get loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix state variables
        flags = fix_state_vars(self, state_args)
        # Check when the state vars are fixed already result in dof 0
        for k in self.keys():
            dof = degrees_of_freedom(self[k])
            if dof != 0:
                raise PropertyPackageError(
                    "\nWhile initializing {sb_name}, the degrees of freedom "
                    "are {dof}, when zero is required. \nInitialization assumes "
                    "that the state variables should be fixed and that no other "
                    "variables are fixed. \nIf other properties have a "
                    "predetermined value, use the calculate_state method "
                    "before using initialize to determine the values for "
                    "the state variables and avoid fixing the property variables."
                    "".format(sb_name=self.name, dof=dof)
                )

        # ---------------------------------------------------------------------
        skip_solve = True  # skip solve if only state variables are present
        for k in self.keys():
            if number_unfixed_variables(self[k]) != 0:
                skip_solve = False

        if not skip_solve:
            # Initialize properties
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = solve_indexed_blocks(opt, [self], tee=slc.tee)
            init_log.info_high(
                "Property initialization: {}.".format(idaeslog.condition(results))
            )

        # If input block, return flags, else release state
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

        if (not skip_solve) and (not check_optimal_termination(results)):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please "
                f"check the output logs for more information."
            )

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        # Unfix state variables
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        revert_state_vars(self, flags)
        init_log.info_high("{} State Released.".format(self.name))

    def calculate_state(
        self,
        var_args=None,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Solves state blocks given a set of variables and their values. These variables can
        be state variables or properties. This method is typically used before
        initialization to solve for state variables because non-state variables (i.e. properties)
        cannot be fixed in initialization routines.

        Keyword Arguments:
            var_args : dictionary with variables and their values, they can be state variables or properties
                       {(VAR_NAME, INDEX): VALUE}
            hold_state : flag indicating whether all of the state variables should be fixed after calculate state.
                         True - State variables will be fixed.
                         False - State variables will remain unfixed, unless already fixed.
            outlvl : idaes logger object that sets output level of solve call (default=idaeslog.NOTSET)
            solver : solver name string if None is provided the default solver
                     for IDAES will be used (default = None)
            optarg : solver options dictionary object (default={})

        Returns:
            results object from state block solve
        """
        # Get logger
        solve_log = idaeslog.getSolveLogger(self.name, level=outlvl, tag="properties")

        # Initialize at current state values (not user provided)
        self.initialize(solver=solver, optarg=optarg, outlvl=outlvl)

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix variables and check degrees of freedom
        flags = (
            {}
        )  # dictionary noting which variables were fixed and their previous state
        for k in self.keys():
            sb = self[k]
            for (v_name, ind), val in var_args.items():
                var = getattr(sb, v_name)
                if iscale.get_scaling_factor(var[ind]) is None:
                    _log.warning(
                        "While using the calculate_state method on {sb_name}, variable {v_name} "
                        "was provided as an argument in var_args, but it does not have a scaling "
                        "factor. This suggests that the calculate_scaling_factor method has not been "
                        "used or the variable was created on demand after the scaling factors were "
                        "calculated. It is recommended to touch all relevant variables (i.e. call "
                        "them or set an initial value) before using the calculate_scaling_factor "
                        "method.".format(v_name=v_name, sb_name=sb.name)
                    )
                if var[ind].is_fixed():
                    flags[(k, v_name, ind)] = True
                    if value(var[ind]) != val:
                        raise ConfigurationError(
                            "While using the calculate_state method on {sb_name}, {v_name} was "
                            "fixed to a value {val}, but it was already fixed to value {val_2}. "
                            "Unfix the variable before calling the calculate_state "
                            "method or update var_args."
                            "".format(
                                sb_name=sb.name,
                                v_name=var.name,
                                val=val,
                                val_2=value(var[ind]),
                            )
                        )
                else:
                    flags[(k, v_name, ind)] = False
                    var[ind].fix(val)

            if degrees_of_freedom(sb) != 0:
                raise RuntimeError(
                    "While using the calculate_state method on {sb_name}, the degrees "
                    "of freedom were {dof}, but 0 is required. Check var_args and ensure "
                    "the correct fixed variables are provided."
                    "".format(sb_name=sb.name, dof=degrees_of_freedom(sb))
                )

        # Solve
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = solve_indexed_blocks(opt, [self], tee=slc.tee)
            solve_log.info_high(
                "Calculate state: {}.".format(idaeslog.condition(results))
            )

        if not check_optimal_termination(results):
            _log.warning(
                "While using the calculate_state method on {sb_name}, the solver failed "
                "to converge to an optimal solution. This suggests that the user provided "
                "infeasible inputs, or that the model is poorly scaled, poorly initialized, "
                "or degenerate."
            )

        # unfix all variables fixed with var_args
        for (k, v_name, ind), previously_fixed in flags.items():
            if not previously_fixed:
                var = getattr(self[k], v_name)
                var[ind].unfix()

        # fix state variables if hold_state
        if hold_state:
            fix_state_vars(self)

        return results


# this third class, provides the instructions to create state blocks
@declare_process_block_class(
    "CoagulationStateBlock", block_class=_CoagulationStateBlock
)
class CoagulationStateBlockData(StateBlockData):
    def build(self):
        """Callable method for Block construction."""
        super(CoagulationStateBlockData, self).build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # first we create the state variables
        #   note: the following dictionaries are used for providing initial values
        #   to state variables and on-demand variables
        self.seawater_mass_frac_dict = {
            ("Liq", "TSS"): 0.01,
            ("Liq", "H2O"): 1,
            ("Liq", "TDS"): 0.01,
            ("Liq", "Sludge"): 0.001,
        }

        self.seawater_mass_flow_dict = {
            ("Liq", "TSS"): 0.01,
            ("Liq", "H2O"): 1,
            ("Liq", "TDS"): 0.01,
            ("Liq", "Sludge"): 0.001,
        }

        self.seawater_mass_conc_dict = {
            ("Liq", "TSS"): 0.01 * 1000,
            ("Liq", "H2O"): 1 * 1000,
            ("Liq", "TDS"): 0.01 * 1000,
            ("Liq", "Sludge"): 0.001 * 1000,
        }

        self.flow_mass_phase_comp = Var(
            self.seawater_mass_frac_dict.keys(),
            initialize=self.seawater_mass_flow_dict,
            bounds=(0.0, None),
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Mass flow rate",
        )

        self.temperature = Var(
            initialize=298.15,
            bounds=(273.15, 700),
            domain=NonNegativeReals,
            units=pyunits.degK,
            doc="State temperature",
        )

        self.pressure = Var(
            initialize=101325,
            bounds=(0.9e5, 3e7),
            domain=NonNegativeReals,
            units=pyunits.Pa,
            doc="State pressure",
        )

    # -----------------------------------------------------------------------------
    # Property Methods
    # these property methods build variables and constraints on demand (i.e. only when asked
    # for by a user or unit model)
    def _mass_frac_phase_comp(self):
        self.mass_frac_phase_comp = Var(
            self.seawater_mass_frac_dict.keys(),
            initialize=self.seawater_mass_frac_dict,
            bounds=(0.0, 1.001),
            units=pyunits.dimensionless,
            doc="Mass fraction",
        )

        def rule_mass_frac_phase_comp(b, p, j):
            return b.mass_frac_phase_comp[p, j] == b.flow_mass_phase_comp[p, j] / (
                b.flow_mass_phase_comp[p, "H2O"]
                + b.flow_mass_phase_comp[p, "TDS"]
                + b.flow_mass_phase_comp[p, "TSS"]
                + b.flow_mass_phase_comp[p, "Sludge"]
            )

        self.eq_mass_frac_phase_comp = Constraint(
            self.seawater_mass_frac_dict.keys(), rule=rule_mass_frac_phase_comp
        )

    def _dens_mass_phase(self):
        self.dens_mass_phase = Var(
            self.params.phase_list,
            initialize={"Liq": 1000},
            bounds=(5e2, 2e4),
            units=pyunits.kg * pyunits.m**-3,
            doc="Mass density",
        )

        #   Density calculation as a function of temperature and pressure
        #   --------------------------------------------------------------
        #   Engineering Toolbox. Water - Density, Specific Weight, and
        #   Thermal Expansion Coefficients. (2003) https://www.engineeringtoolbox.com/
        #   water-density-specific-weight-d_595.html [Accessed 02-01-2022]
        def rule_dens_mass_phase(b, p):
            return b.dens_mass_phase[p] == (
                b.params.ref_dens_liq
                + b.params.dens_slope * b.mass_frac_phase_comp[p, "TDS"]
                + b.params.dens_slope * b.mass_frac_phase_comp[p, "TSS"]
                + b.params.dens_slope * b.mass_frac_phase_comp[p, "Sludge"]
            ) * (
                b.params.dens_param_A * b.temperature**2
                + b.params.dens_param_B * b.temperature
                + b.params.dens_param_C
            ) * (
                b.params.ref_pressure_correction
                + b.params.ref_pressure_slope * b.pressure
            )

        self.eq_dens_mass_phase = Constraint(
            self.params.phase_list, rule=rule_dens_mass_phase
        )

    def _flow_vol_phase(self):
        self.flow_vol_phase = Var(
            self.params.phase_list,
            initialize={"Liq": 1e-3},
            bounds=(0.0, None),
            units=pyunits.m**3 / pyunits.s,
            doc="Volumetric flow rate",
        )

        def rule_flow_vol_phase(b, p):
            return (
                b.flow_vol_phase[p]
                == (
                    b.flow_mass_phase_comp[p, "H2O"]
                    + b.flow_mass_phase_comp[p, "TDS"]
                    + b.flow_mass_phase_comp[p, "TSS"]
                    + b.flow_mass_phase_comp[p, "Sludge"]
                )
                / b.dens_mass_phase[p]
            )

        self.eq_flow_vol_phase = Constraint(
            self.params.phase_list, rule=rule_flow_vol_phase
        )

    def _conc_mass_phase_comp(self):
        self.conc_mass_phase_comp = Var(
            self.seawater_mass_frac_dict.keys(),
            initialize=self.seawater_mass_conc_dict,
            bounds=(0.0, None),
            units=pyunits.kg / pyunits.m**3,
            doc="Mass concentration",
        )

        def rule_conc_mass_phase_comp(b, p, j):
            return (
                b.conc_mass_phase_comp[p, j]
                == b.mass_frac_phase_comp[p, j] * b.dens_mass_phase[p]
            )

        self.eq_conc_mass_phase_comp = Constraint(
            self.seawater_mass_frac_dict.keys(), rule=rule_conc_mass_phase_comp
        )

    def _visc_d_phase(self):
        self.visc_d_phase = Var(
            self.params.phase_list,
            initialize={"Liq": 0.001},
            bounds=(0.0, 0.01),
            units=pyunits.kg / pyunits.m / pyunits.s,
            doc="Dynamic viscosity",
        )

        #   Adjustment for dynamic viscosity as a function of temperature
        #   -------------------------------------------------------------
        #   D.S. Viswananth, G. Natarajan. Data Book on the Viscosity of
        #     Liquids. Hemisphere Publishing Corp. (1989)
        def rule_visc_d_phase(b, p):
            return b.visc_d_phase[p] == (
                b.params.mu_A * exp(b.params.mu_B / (b.temperature - b.params.mu_C))
            )

        self.eq_visc_d_phase = Constraint(
            self.params.phase_list, rule=rule_visc_d_phase
        )

    def _enth_flow(self):
        # enthalpy flow expression for get_enthalpy_flow_terms method
        temperature_ref = 273.15 * pyunits.K

        def rule_enth_flow(b, p):  # enthalpy flow [J/s]
            return (
                b.params.cp
                * sum(b.flow_mass_phase_comp[pair] for pair in b.flow_mass_phase_comp)
                * (b.temperature - temperature_ref)
            )

        self.enth_flow = Expression(self.params.phase_list, rule=rule_enth_flow)

    # -----------------------------------------------------------------------------
    # General Methods
    # NOTE: For scaling in the control volume to work properly, these methods must
    # return a pyomo Var or Expression

    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        return self.flow_mass_phase_comp[p, j]

    def get_enthalpy_flow_terms(self, p):
        """Create enthalpy flow terms."""
        return self.enth_flow[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass

    def define_state_vars(self):
        """Define state vars."""
        return {
            "flow_mass_phase_comp": self.flow_mass_phase_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    # -----------------------------------------------------------------------------
    # Scaling methods
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # scaling factors for parameters
        if iscale.get_scaling_factor(self.params.cp) is None:
            iscale.set_scaling_factor(self.params.cp, 1e-3)

        # these variables should have user input
        if iscale.get_scaling_factor(self.flow_mass_phase_comp["Liq", "H2O"]) is None:
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Liq", "H2O"], default=1, warning=True
            )
            iscale.set_scaling_factor(self.flow_mass_phase_comp["Liq", "H2O"], sf)

        if iscale.get_scaling_factor(self.flow_mass_phase_comp["Liq", "TSS"]) is None:
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Liq", "TSS"], default=1e2, warning=True
            )
            iscale.set_scaling_factor(self.flow_mass_phase_comp["Liq", "TSS"], sf)

        if iscale.get_scaling_factor(self.flow_mass_phase_comp["Liq", "TDS"]) is None:
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Liq", "TDS"], default=1e2, warning=True
            )
            iscale.set_scaling_factor(self.flow_mass_phase_comp["Liq", "TDS"], sf)

        if (
            iscale.get_scaling_factor(self.flow_mass_phase_comp["Liq", "Sludge"])
            is None
        ):
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Liq", "Sludge"], default=1e3, warning=True
            )
            iscale.set_scaling_factor(self.flow_mass_phase_comp["Liq", "Sludge"], sf)

        # these variables do not typically require user input,
        # will not override if the user does provide the scaling factor

        if self.is_property_constructed("mass_frac_phase_comp"):
            # Apply variable scaling
            for pair in self.seawater_mass_frac_dict.keys():
                if iscale.get_scaling_factor(self.mass_frac_phase_comp[pair]) is None:
                    if pair[1] == "H2O":
                        iscale.set_scaling_factor(self.mass_frac_phase_comp[pair], 1)
                    else:
                        sf = iscale.get_scaling_factor(
                            self.flow_mass_phase_comp[pair]
                        ) / iscale.get_scaling_factor(
                            self.flow_mass_phase_comp["Liq", "H2O"]
                        )
                        iscale.set_scaling_factor(self.mass_frac_phase_comp[pair], sf)

            # Apply constraint scaling
            for pair in self.seawater_mass_frac_dict.keys():
                sf = iscale.get_scaling_factor(
                    self.mass_frac_phase_comp[pair], default=1, warning=True
                )
                iscale.constraint_scaling_transform(
                    self.eq_mass_frac_phase_comp[pair], sf
                )

        if self.is_property_constructed("dens_mass_phase"):
            if iscale.get_scaling_factor(self.dens_mass_phase) is None:
                iscale.set_scaling_factor(self.dens_mass_phase, 1e-3)

        if self.is_property_constructed("flow_vol_phase"):
            sf = iscale.get_scaling_factor(
                self.flow_mass_phase_comp["Liq", "H2O"]
            ) / iscale.get_scaling_factor(self.dens_mass_phase["Liq"])
            iscale.set_scaling_factor(self.flow_vol_phase["Liq"], sf)

        if self.is_property_constructed("conc_mass_phase_comp"):
            # Apply variable scaling
            for pair in self.seawater_mass_frac_dict.keys():
                if iscale.get_scaling_factor(self.conc_mass_phase_comp[pair]) is None:
                    if pair[0] == "Liq":
                        sf = iscale.get_scaling_factor(
                            self.mass_frac_phase_comp[pair]
                        ) * iscale.get_scaling_factor(self.dens_mass_phase["Liq"])
                        iscale.set_scaling_factor(self.conc_mass_phase_comp[pair], sf)
                    else:
                        raise PropertyPackageError(
                            "Unsupported phase for CoagulationParameterData property package"
                        )

            # Apply constraint scaling
            for pair in self.seawater_mass_frac_dict.keys():
                sf = iscale.get_scaling_factor(
                    self.conc_mass_phase_comp[pair], default=1, warning=True
                )
                iscale.constraint_scaling_transform(
                    self.eq_conc_mass_phase_comp[pair], sf
                )

        if self.is_property_constructed("visc_d_phase"):
            if iscale.get_scaling_factor(self.visc_d_phase) is None:
                iscale.set_scaling_factor(self.visc_d_phase, 1e3)

        if self.is_property_constructed("enth_flow"):
            if iscale.get_scaling_factor(self.enth_flow) is None:
                sf = (
                    iscale.get_scaling_factor(self.params.cp)
                    * iscale.get_scaling_factor(self.flow_mass_phase_comp["Liq", "H2O"])
                    * 1e-1
                )
                iscale.set_scaling_factor(self.enth_flow, sf)

        # transforming constraints
        transform_property_constraints(self)

    # ------------------------------------------------------------------
    # # TODO: Create functions for easier user access to setup intial values
    #       for all state vars and properties. This may be useful for improving
    #       how scaling factors are determined.



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
import watertap.property_models.cryst_prop_pack as props
from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlock, ControlVolume0DBlock
from idaes.models.properties.tests.test_harness import (
    PropertyTestHarness as PropertyTestHarness_idaes,
)
from watertap.property_models.tests.property_test_harness import (
    PropertyTestHarness,
    PropertyRegressionTest,
    PropertyCalculateStateTest,
)


# -----------------------------------------------------------------------------


class TestNaClProperty_idaes(PropertyTestHarness_idaes):
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}
        self.prop_args = {}
        self.has_density_terms = False


class TestDefaultNaClwaterProperty:

    # Create block and stream for running default tests
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = props.NaClParameterBlock(
        heat_of_crystallization_model=props.HeatOfCrystallizationModel.constant
    )
    m.fs.stream = m.fs.properties.build_state_block([0], defined_state=True)

    m.fs.cv = ControlVolume0DBlock(
        dynamic=False, has_holdup=False, property_package=m.fs.properties
    )
    m.fs.cv.add_state_blocks(has_phase_equilibrium=False)
    m.fs.cv.add_material_balances()
    m.fs.cv.add_energy_balances()
    m.fs.cv.add_momentum_balances()

    # Create instance of PropertyTesthARNESS class and add attributes needed for tests
    xv = PropertyTestHarness()

    xv.stateblock_statistics = {
        "number_variables": 43,
        "number_total_constraints": 37,
        "number_unused_variables": 0,
        "default_degrees_of_freedom": 6,
    }

    xv.scaling_args = {
        ("flow_mass_phase_comp", ("Liq", "H2O")): 1e0,
        ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e1,
        ("flow_mol_phase_comp", ("Sol", "NaCl")): 1e2,
        ("flow_mol_phase_comp", ("Vap", "H2O")): 1e3,
    }

    xv.default_solution = {
        ("solubility_mass_phase_comp", ("Liq", "NaCl")): 359.51,
        ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.265,
        ("dens_mass_solvent", "Liq"): 996.89,
        ("dens_mass_solvent", "Vap"): 0.7363,
        ("dens_mass_solute", "Liq"): 3199.471,
        ("dens_mass_solute", "Sol"): 2115,
        ("dens_mass_phase", "Liq"): 1021.50,
        ("dh_vap_mass_solvent", None): 2441.80,
        ("cp_mass_solvent", "Liq"): 4186.52,
        ("cp_mass_solvent", "Vap"): 1864.52,
        ("cp_mass_solute", "Sol"): 864.15,  # cp_mass_solute liquid ignored for now
        ("cp_mass_phase", "Liq"): 4008.30,
        ("flow_vol_phase", "Liq"): 9.79e-4,
        ("flow_vol_phase", "Sol"): 0,
        ("flow_vol_phase", "Vap"): 0,
        ("pressure_sat", None): 2932.43,
        ("temperature_sat_solvent", None): 296.79,
        ("conc_mass_phase_comp", ("Liq", "NaCl")): 35.753,
        ("conc_mass_phase_comp", ("Liq", "H2O")): 985.753,
        ("enth_mass_solvent", "Liq"): 104.9212,
        ("enth_mass_solvent", "Vap"): 2546.7289,
        ("enth_mass_solute", "Sol"): 21.4739,
        ("enth_mass_phase", "Liq"): 101.0922,
        ("dh_crystallization_mass_comp", "NaCl"): -520,
        ("mass_frac_phase_comp", ("Liq", "H2O")): 0.965,
        ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.035,
        ("mass_frac_phase_comp", ("Sol", "NaCl")): 1.0,
        ("mass_frac_phase_comp", ("Vap", "H2O")): 1.0,
        ("flow_mol_phase_comp", ("Liq", "H2O")): 53.57,
        ("flow_mol_phase_comp", ("Liq", "NaCl")): 0.5989,
        ("flow_mol_phase_comp", ("Sol", "NaCl")): 0.0,
        ("flow_mol_phase_comp", ("Vap", "H2O")): 0.0,
        ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9889,
        ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.01106,
        ("mole_frac_phase_comp", ("Sol", "NaCl")): 1.0,
        ("mole_frac_phase_comp", ("Vap", "H2O")): 1.0,
    }

    # Configure class
    xv.configure_class(m)

    @pytest.mark.unit
    def test_components_phases(self):
        self.xv.test_components_phases(self.m)

    @pytest.mark.unit
    def test_parameters(self):
        self.xv.test_parameters(self.m)

    @pytest.mark.unit
    def test_state_variables(self):
        self.xv.test_state_variables(self.m)

    @pytest.mark.unit
    def test_permanent_properties(self):
        self.xv.test_permanent_properties(self.m)

    @pytest.mark.unit
    def test_on_demand_properties(self):
        self.xv.test_on_demand_properties(self.m)

    @pytest.mark.unit
    def test_stateblock_statistics(self):
        self.xv.test_stateblock_statistics(self.m)

    @pytest.mark.unit
    def test_units_consistent(self):
        self.xv.test_units_consistent(self.m)

    @pytest.mark.unit
    def test_scaling(self):
        self.xv.test_scaling(self.m)

    @pytest.mark.component
    def test_default_initialization(self):
        self.xv.test_default_initialization(self.m)

    @pytest.mark.component
    def test_property_control_volume(self):
        self.xv.test_property_control_volume(self.m)


@pytest.mark.component
class TestNaClPropertySolution_1(PropertyRegressionTest):
    # Test pure liquid solution 1 - same solution as NaCl prop pack
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e2,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.95,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.05,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0.0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0.0,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 50e5,
        }

        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.95,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 359.50,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.265,
            ("dens_mass_solvent", "Liq"): 996.89,
            ("dens_mass_solvent", "Vap"): 36.335,
            ("dens_mass_phase", "Liq"): 1032.2,
            ("flow_vol_phase", "Liq"): 9.687e-4,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 980.6,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 51.61,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 52.73,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 0.8556,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9840,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 1.597e-2,
            ("cp_mass_solvent", "Liq"): 4186.52,
            ("cp_mass_phase", "Liq"): 3940.03,
            ("enth_mass_solvent", "Liq"): 104.92,
            ("enth_mass_phase", "Liq"): 99.45,
        }


@pytest.mark.component
class TestNaClPropertySolution_2(PropertyRegressionTest):
    # Test pure liquid solution 2 - same solution as NaCl prop pack
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e2,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.74,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.26,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0.0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0.0,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 50e5,
        }

        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.74,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.26,
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 359.50,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.265,
            ("dens_mass_solvent", "Liq"): 996.89,
            ("dens_mass_phase", "Liq"): 1193.4,
            ("flow_vol_phase", "Liq"): 8.379e-4,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 883.14,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 310.29,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 41.08,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 4.449,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9022,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 9.773e-2,
            ("cp_mass_solvent", "Liq"): 4186.52,
            ("cp_mass_phase", "Liq"): 3276.56,
            ("enth_mass_solvent", "Liq"): 104.92,
            ("enth_mass_phase", "Liq"): 68.78,
        }


@pytest.mark.component
class TestNaClPropertySolution_3(PropertyRegressionTest):
    # Test pure liquid solution 3 - same solution as NaCl prop pack
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e3,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e3,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.999,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.001,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0.0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0.0,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 10e5,
        }

        self.regression_solution = {
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.999,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.001,
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 359.50,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.265,
            ("dens_mass_solvent", "Liq"): 996.89,
            ("dens_mass_solvent", "Vap"): 7.267,
            ("dens_mass_phase", "Liq"): 997.59,
            ("flow_vol_phase", "Liq"): 1.002e-3,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 996.59,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 0.9976,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 55.45,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 1.711e-2,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9997,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 3.084e-4,
            ("cp_mass_solvent", "Liq"): 4186.52,
            ("cp_mass_phase", "Liq"): 4180.98,
            ("enth_mass_solvent", "Liq"): 104.92,
            ("enth_mass_phase", "Liq"): 104.40,
        }


@pytest.mark.requires_idaes_solver
@pytest.mark.component
class TestNaClPropertySolution_4(PropertyRegressionTest):
    # Test pure solid solution 1 - check solid properties
    def configure(self):

        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e3,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e2,
        }

        self.state_args = {
            ("flow_vol_phase", "Liq"): 0,
            ("flow_vol_phase", "Vap"): 0,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1.0,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }

        self.regression_solution = {
            ("mass_frac_phase_comp", ("Sol", "NaCl")): 1.0,
            ("dens_mass_solute", "Sol"): 2115,
            ("cp_mass_solute", "Sol"): 864.16,
            ("flow_vol_phase", "Sol"): 1 / 2115,  # mass floe / density
            ("cp_mass_solute", "Sol"): 864.16,
            ("enth_mass_solute", "Sol"): 21.474,
            ("dh_crystallization_mass_comp", "NaCl"): -520,
            ("flow_mol_phase_comp", ("Sol", "NaCl")): 1 / 58.44e-3,  # mass flow / mw
            ("mole_frac_phase_comp", ("Sol", "NaCl")): 1.0,
        }


@pytest.mark.requires_idaes_solver
@pytest.mark.component
class TestNaClPropertySolution_5(PropertyRegressionTest):
    # Test pure vapor solution 1 - check vapor properties
    def configure(self):

        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e3,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e3,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e0,
        }

        self.state_args = {
            ("flow_vol_phase", "Liq"): 0,
            ("flow_vol_phase", "Sol"): 0,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1.0,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }

        self.regression_solution = {
            ("mass_frac_phase_comp", ("Vap", "H2O")): 1.0,
            ("dens_mass_solvent", "Vap"): 3.633,
            ("dh_vap_mass_solvent", None): 2441.808,
            ("cp_mass_solvent", "Vap"): 1864.52,
            ("flow_vol_phase", "Vap"): 1 / 3.633,  # mass flow / density
            ("pressure_sat", None): 2905.28,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 1 / 18.01528e-3,  # mass flow / mw
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1.0,
        }


@pytest.mark.component
class TestNaClPropertySolution_6(PropertyRegressionTest):
    # Test for S-L-V solution 1 with similar magnitude flowrates in all phases and high liquid salt. conc. - check all properties
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 10,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e1,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.75,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.25,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0.25,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0.25,
            ("temperature", None): 273.15 + 50,
            ("pressure", None): 5e5,
        }

        self.regression_solution = {
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 362.93,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.2686,
            ("dens_mass_solvent", "Liq"): 988.04,
            ("dens_mass_solvent", "Vap"): 3.352,
            ("dens_mass_solute", "Liq"): 2645.21,
            ("dens_mass_solute", "Sol"): 2115,
            ("dens_mass_phase", "Liq"): 1171.53,
            ("dh_vap_mass_solvent", None): 2382.08,
            ("cp_mass_solvent", "Liq"): 4180.92,
            ("cp_mass_solvent", "Vap"): 1871.21,
            ("cp_mass_solute", "Sol"): 873.34,
            ("cp_mass_phase", "Liq"): 3300.83,
            ("flow_vol_phase", "Liq"): 8.53e-4,
            ("flow_vol_phase", "Sol"): 1.182e-4,
            ("flow_vol_phase", "Vap"): 7.46e-2,
            ("pressure_sat", None): 9799.91,
            ("temperature_sat_solvent", None): 318.52,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 292.88,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 878.65,
            ("enth_mass_solvent", "Liq"): 209.40,
            ("enth_mass_solvent", "Vap"): 2591.48,
            ("enth_mass_solute", "Sol"): 43.19,
            ("enth_mass_phase", "Liq"): 152.36,
            ("dh_crystallization_mass_comp", "NaCl"): -520,
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.75,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.25,
            ("mass_frac_phase_comp", ("Sol", "NaCl")): 1,
            ("mass_frac_phase_comp", ("Vap", "H2O")): 1,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 41.63,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 4.28,
            ("flow_mol_phase_comp", ("Sol", "NaCl")): 4.28,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 13.88,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9068,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.0932,
            ("mole_frac_phase_comp", ("Sol", "NaCl")): 1.0,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1.0,
        }


@pytest.mark.component
class TestNaClPropertySolution_7(PropertyRegressionTest):
    # Test for S-L-V solution 2 with flowrates in all phases of same magnitude but low liquid salt. conc. - check all properties
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 10,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e3,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0.999,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0.001,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0.25,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0.25,
            ("temperature", None): 273.15 + 50,
            ("pressure", None): 5e5,
        }

        self.regression_solution = {
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 362.93,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.2686,
            ("dens_mass_solvent", "Liq"): 988.04,
            ("dens_mass_solvent", "Vap"): 3.352,
            ("dens_mass_solute", "Liq"): 3073.05,
            ("dens_mass_solute", "Sol"): 2115,
            ("dens_mass_phase", "Liq"): 988.72,
            ("dh_vap_mass_solvent", None): 2382.08,
            ("cp_mass_solvent", "Liq"): 4180.92,
            ("cp_mass_solvent", "Vap"): 1871.21,
            ("cp_mass_solute", "Sol"): 873.34,
            ("cp_mass_phase", "Liq"): 4175.95,
            ("flow_vol_phase", "Liq"): 1.011e-3,
            ("flow_vol_phase", "Sol"): 1.182e-4,
            ("flow_vol_phase", "Vap"): 7.46e-2,
            ("pressure_sat", None): 12614.93,
            ("temperature_sat_solvent", None): 323.55,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 0.9887,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 987.73,
            ("enth_mass_solvent", "Liq"): 209.40,
            ("enth_mass_solvent", "Vap"): 2591.48,
            ("enth_mass_solute", "Sol"): 43.19,
            ("enth_mass_phase", "Liq"): 208.81,
            ("dh_crystallization_mass_comp", "NaCl"): -520,
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.999,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.001,
            ("mass_frac_phase_comp", ("Sol", "NaCl")): 1,
            ("mass_frac_phase_comp", ("Vap", "H2O")): 1,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 55.45,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 0.0171,
            ("flow_mol_phase_comp", ("Sol", "NaCl")): 4.28,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 13.88,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.999,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 3.084e-4,
            ("mole_frac_phase_comp", ("Sol", "NaCl")): 1.0,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1.0,
        }


@pytest.mark.component
class TestNaClPropertySolution_8(PropertyRegressionTest):
    # Test for S-L-V solution 3 with outlet data from Dutta et al. - proper crystallization system
    # # Dutta recorded properties for solids and liquids at crystallizer temperature:
    # # Solubility @ 55C: 0.27 kg/kg
    # # Heat of vaporization @ 55C: 2400 kJ/kg
    # # Vapor density @ 55C: 68.7e-3 kg/m3
    # # Solid heat capacity: 877 J/kgK
    # # Liquid heat capacity @ 20 C: 3290 kJ/kgK
    # # Liquid density @ 20 C : 1185 kg/m3

    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e-1,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 18.37,  # 84t/h
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.2126,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 5.55,  # 20t/h
            ("flow_mass_phase_comp", ("Vap", "H2O")): 20.55,  # 74t/h
            ("temperature", None): 273.15 + 20,
            ("pressure", None): 10e3,
        }

        self.regression_solution = {
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 358.88,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.2644,
            ("dens_mass_solvent", "Liq"): 998.02,
            ("dens_mass_solvent", "Vap"): 73.91e-3,
            ("dens_mass_solute", "Liq"): 2847.63,
            ("dens_mass_solute", "Sol"): 2115,
            ("dens_mass_phase", "Liq"): 1157.91,
            ("dh_vap_mass_solvent", None): 2453.66,
            ("cp_mass_solvent", "Liq"): 4189.43,
            ("cp_mass_solvent", "Vap"): 1863.46,
            ("cp_mass_solute", "Sol"): 862.16,
            ("cp_mass_phase", "Liq"): 3380.07,
            ("flow_vol_phase", "Liq"): 0.02015,
            ("flow_vol_phase", "Sol"): 2.624e-3,
            ("flow_vol_phase", "Vap"): 278.04,
            ("pressure_sat", None): 1679.64,
            ("temperature_sat_solvent", None): 287.88,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 246.17,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 911.74,
            ("enth_mass_solvent", "Liq"): 84.00,
            ("enth_mass_solvent", "Vap"): 2537.66,
            ("enth_mass_solute", "Sol"): 17.16,
            ("enth_mass_phase", "Liq"): 59.03,
            ("dh_crystallization_mass_comp", "NaCl"): -520,
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.7874,
            ("mass_frac_phase_comp", ("Sol", "NaCl")): 1,
            ("mass_frac_phase_comp", ("Vap", "H2O")): 1,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 1019.69,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 84.88,
            ("flow_mol_phase_comp", ("Sol", "NaCl")): 94.97,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 1140.698,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9232,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.0768,
            ("mole_frac_phase_comp", ("Sol", "NaCl")): 1.0,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1.0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 4.96,
        }


@pytest.mark.component
class TestNaClPropertySolution_9(PropertyRegressionTest):
    # Test for S-L-V solution 4 with outlet data from Dutta et al. - proper crystallization system
    # # Dutta recorded properties for solids and liquids at crystallizer temperature:
    # # Solubility @ 55C: 0.27 kg/kg
    # # Heat of vaporization @ 55C: 2400 kJ/kg
    # # Vapor density @ 55C: 68.7e-3 kg/m3
    # # Solid heat capacity: 877 J/kgK
    # # Liquid heat capacity @ 20 C: 3290 kJ/kgK
    # # Liquid density @ 20 C : 1185 kg/m3

    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e-1,
        }

        self.state_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 18.37,  # 84t/h
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.2126,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 5.55,  # 20t/h
            ("flow_mass_phase_comp", ("Vap", "H2O")): 20.55,  # 74t/h
            ("temperature", None): 273.15 + 55,
            ("pressure", None): 10e3,
        }

        self.regression_solution = {
            ("solubility_mass_phase_comp", ("Liq", "NaCl")): 363.71,
            ("solubility_mass_frac_phase_comp", ("Liq", "NaCl")): 0.2695,
            ("dens_mass_solvent", "Liq"): 985.71,
            ("dens_mass_solvent", "Vap"): 66.03e-3,
            ("dens_mass_solute", "Liq"): 2694.61,
            ("dens_mass_solute", "Sol"): 2115,
            ("dens_mass_phase", "Liq"): 1139.33,
            ("dh_vap_mass_solvent", None): 2369.98,
            ("cp_mass_solvent", "Liq"): 4181.59,
            ("cp_mass_solvent", "Vap"): 1872.79,
            ("cp_mass_solute", "Sol"): 875.04,
            ("cp_mass_phase", "Liq"): 3390.43,
            ("flow_vol_phase", "Liq"): 0.02047,
            ("flow_vol_phase", "Sol"): 2.624e-3,
            ("flow_vol_phase", "Vap"): 311.233,
            ("pressure_sat", None): 13201.77,
            ("temperature_sat_solvent", None): 324.47,
            ("conc_mass_phase_comp", ("Liq", "NaCl")): 242.22,
            ("conc_mass_phase_comp", ("Liq", "H2O")): 897.107,
            ("enth_mass_solvent", "Liq"): 230.30,
            ("enth_mass_solvent", "Vap"): 2600.279,
            ("enth_mass_solute", "Sol"): 47.57,
            ("enth_mass_phase", "Liq"): 177.39,
            ("dh_crystallization_mass_comp", "NaCl"): -520,
            ("mass_frac_phase_comp", ("Liq", "H2O")): 0.7874,
            ("mass_frac_phase_comp", ("Sol", "NaCl")): 1,
            ("mass_frac_phase_comp", ("Vap", "H2O")): 1,
            ("flow_mol_phase_comp", ("Liq", "H2O")): 1019.69,
            ("flow_mol_phase_comp", ("Liq", "NaCl")): 84.88,
            ("flow_mol_phase_comp", ("Sol", "NaCl")): 94.97,
            ("flow_mol_phase_comp", ("Vap", "H2O")): 1140.698,
            ("mole_frac_phase_comp", ("Liq", "H2O")): 0.9232,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.0768,
            ("mole_frac_phase_comp", ("Sol", "NaCl")): 1.0,
            ("mole_frac_phase_comp", ("Vap", "H2O")): 1.0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 4.96,
        }


@pytest.mark.component
class TestNaClCalculateState_1(PropertyCalculateStateTest):
    # Test pure liquid solution with mass fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1,
        }

        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-2,
            ("flow_vol_phase", "Sol"): 0,
            ("flow_vol_phase", "Vap"): 0,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }

        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 19.6,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1.032,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0,
        }


@pytest.mark.component
class TestNaClCalculateState_2(PropertyCalculateStateTest):
    # Test pure liquid with mole fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
            # The rest are expected to be zero
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1,
        }
        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-4,
            ("flow_vol_phase", "Sol"): 0,
            ("flow_vol_phase", "Vap"): 0,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }
        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 18.84e-2,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 3.215e-2,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0,
        }


@pytest.mark.component
class TestNaClCalculateState_3(PropertyCalculateStateTest):
    # Test pure liquid solution with pressure_sat defined instead of temperature
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e1,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e-1,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e2,
        }

        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-2,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("flow_vol_phase", "Sol"): 0,
            ("flow_vol_phase", "Vap"): 0,
            ("pressure_sat", None): 2905,
            ("pressure", None): 5e5,
        }

        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 19.6,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1.032,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0,
            ("temperature", None): 273.15 + 25,
        }


@pytest.mark.component
class TestNaClCalculateState_4(PropertyCalculateStateTest):
    # Test pure solid solution with mass fractions
    def configure(self):

        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e-1,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e-1,
        }

        self.var_args = {
            ("flow_vol_phase", "Liq"): 0,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0,
            ("flow_vol_phase", "Sol"): 2e-2,
            ("flow_vol_phase", "Vap"): 0,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }

        self.state_solution = {
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 2115
            * 2e-2,  # solid density is constant
            ("flow_mass_phase_comp", ("Liq", "H2O")): 0,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 0,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0,
        }


@pytest.mark.component
class TestNaClCalculateState_5(PropertyCalculateStateTest):
    # Test solid-liquid-vapor mixture solution with mass fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e-1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e1,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e-2,
        }

        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-2,
            ("mass_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("flow_vol_phase", "Sol"): 2e-2,
            ("flow_vol_phase", "Vap"): 2e-2,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }

        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 19.6,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1.032,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 0.07265,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 42.3,
        }


@pytest.mark.component
class TestNaClCalculateState_6(PropertyCalculateStateTest):
    # Test liquid-solid-vapor mixture with mole fractions
    def configure(self):
        self.prop_pack = props.NaClParameterBlock
        self.param_args = {}

        self.scaling_args = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1e1,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 1e2,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 1e2,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 1e3,
        }

        self.var_args = {
            ("flow_vol_phase", "Liq"): 2e-4,
            ("flow_vol_phase", "Sol"): 2e-4,
            ("flow_vol_phase", "Vap"): 2e-4,
            ("mole_frac_phase_comp", ("Liq", "NaCl")): 0.05,
            ("temperature", None): 273.15 + 25,
            ("pressure", None): 5e5,
        }

        self.state_solution = {
            ("flow_mass_phase_comp", ("Liq", "H2O")): 18.84e-2,
            ("flow_mass_phase_comp", ("Liq", "NaCl")): 3.215e-2,
            ("flow_mass_phase_comp", ("Sol", "NaCl")): 0.423,
            ("flow_mass_phase_comp", ("Vap", "H2O")): 3.632
            * 2e-4,  # Density from ideal gas law * vol. flow
        }



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for loading water source definitions
"""
import pytest
import os

from pyomo.environ import units
from pyomo.util.check_units import assert_units_equivalent

from watertap.core import Database

dbpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")

db = Database()

exclude_files = ["water_sources.yaml", "component_list.yaml", "default_case_study.yaml"]

tech_list = []
for f in os.listdir(dbpath):
    filename = os.fsdecode(f)
    if filename.endswith(".yaml") and filename not in exclude_files:
        tech_list.append(filename[:-5])


@pytest.mark.integration
@pytest.mark.parametrize("tech", tech_list)
def test_unit_parameter_files(tech):
    data = db._get_technology(tech)

    # Check that data has as default key
    assert "default" in data

    # Iterate overall entries in tech data and check for expected contents
    # TODO : Need to check up on this once everything is done
    pass_through = [
        "blending_reservoir",
        "buffer_tank",
        "chemical_addition",
        "co2_addition",
        "cooling_supply",
        "energy_recovery",
        "feed_water_tank",
        "injection_well_disposal",
        "intrusion_mitigation",
        "landfill",
        "municipal_drinking",
        "municipal_wwtp",
        "pump_electricity",
        "pump",
        "smp",
        "static_mixer",
        "storage_tank",
        "surface_discharge",
        "sw_onshore_intake",
        "tramp_oil_tank",
        "water_pumping_station",
        "well_field",
    ]

    siso_full_recovery = ["uv_aop", "uv", "fixed_bed", "decarbonator", "chlorination"]

    no_energy_electric_flow_vol_inlet = [
        "anaerobic_mbr_mec",
        "autothermal_hydrothermal_liquefaction",
        "backwash_solids_handling",
        "brine_concentrator",
        "CANDO_P",
        "chemical_addition",
        "coag_and_floc",
        "cofermentation",
        "constructed_wetlands",
        "deep_well_injection",
        "electrochemical_nutrient_removal",
        "electrodialysis_reversal",
        "electrocoagulation",
        "energy_recovery",
        "evaporation_pond",
        "filter_press",
        "gac",
        "gas_sparged_membrane",
        "hydrothermal_gasification",
        "ion_exchange",
        "iron_and_manganese_removal",
        "mbr",
        "metab",
        "municipal_drinking",
        "ozonation",
        "ozone_aop",
        "photothermal_membrane",
        "pump_electricity",
        "storage_tank",
        "supercritical_salt_precipitation",
        "surface_discharge",
        "sw_onshore_intake",
        "vfa_recovery",
        "water_pumping_station",
        "well_field",
    ]

    expected = ["recovery_frac_mass_H2O", "default_removal_frac_mass_comp"]

    for k in data.values():

        for e in expected:
            if tech not in pass_through and tech not in siso_full_recovery:
                assert e in k.keys()
                assert "units" in k[e].keys()
                assert_units_equivalent(k[e]["units"], units.dimensionless)
                assert "value" in k[e].keys()
                assert k[e]["value"] >= 0
                assert k[e]["value"] <= 1
            elif tech in siso_full_recovery:
                if e == "default_removal_frac_mass_comp":
                    assert e in k.keys()
            else:
                assert e not in k.keys()

        if tech not in no_energy_electric_flow_vol_inlet:
            e = "energy_electric_flow_vol_inlet"
            assert e in k.keys()
            assert "units" in k[e].keys()
            assert_units_equivalent(k[e]["units"], units.dimensionless)
            assert "value" in k[e].keys()
            assert k[e]["value"] >= 0

        # Check for specific removal fractions
        if "removal_frac_mass_comp" in k.keys():
            for (j, c_data) in k["removal_frac_mass_comp"].items():
                assert "units" in c_data.keys()
                assert_units_equivalent(c_data["units"], units.dimensionless)
                assert "value" in c_data.keys()
                assert c_data["value"] >= 0
                assert c_data["value"] <= 1
                assert j in db.component_list.keys()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

import pytest
from pyomo.environ import units as pyunits
from watertap.tools.oli_api.util.state_block_helper_functions import (
    create_state_block,
    extract_state_vars,
)


@pytest.fixture
def chemistry_source():
    return {
        "temperature": 298.15,  # temperature in K
        "pressure": 101325,  # pressure in Pa
        "components": {  # concentrations in mg/L
            "Cl_-": 870,
            "Na_+": 739,
            "SO4_2-": 1011,
            "Mg_2+": 90,
            "Ca_2+": 258,
            "K_+": 9,
            "HCO3_-": 385,
        },
        "units": {
            "temperature": pyunits.K,
            "pressure": pyunits.Pa,
            "components": pyunits.mg / pyunits.L,
        },
    }


@pytest.fixture
def state_block(chemistry_source):
    return create_state_block(chemistry_source)


@pytest.mark.unit
def test_extract_state_vars(chemistry_source, state_block):
    conc_var = state_block.fs.stream[0].conc_mass_phase_comp
    units = chemistry_source["units"]
    extract_state_vars(state_block.fs.stream[0], conc_var, chemistry_source["units"])



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import ConcreteModel, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom

import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

import watertap.examples.custom_model_demo.simple_prop_pack as props
from watertap.examples.custom_model_demo.simple_filter import Filtration


def main():
    # create model, flowsheet
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    # attach property package
    m.fs.properties = props.PropParameterBlock()
    # build the unit model
    m.fs.filter = Filtration(property_package=m.fs.properties)

    # display model
    # note that there are the recovery and removal fraction variables are on m.fs.filter
    # any variable that starts with a _ can be ignored, they are references
    # note that are three separate state blocks on the model (properties_in, properties_out, properties_waste)
    print("first display")
    m.fs.filter.display()

    # now specify the model
    # note there are 8 degrees of freedom
    print("DOF before specifying:", degrees_of_freedom(m.fs))

    # specify the feed
    m.fs.filter.properties_in[0].pressure.fix(2e5)
    m.fs.filter.properties_in[0].temperature.fix(273.15 + 25)
    m.fs.filter.properties_in[0].flow_mass_phase_comp["Liq", "H2O"].fix(1)
    m.fs.filter.properties_in[0].flow_mass_phase_comp["Liq", "NaCl"].fix(0.035)
    m.fs.filter.properties_in[0].flow_mass_phase_comp["Liq", "TSS"].fix(120e-6)
    # an alternative to setting the state variables at the state block is to use the port like below
    # note that the time domain 0, is now accessed with the other indices, this is the case for ports
    # m.fs.filter.inlet.pressure[0].fix(2e5)
    # m.fs.filter.inlet.temperature[0].fix(273.15 + 25)
    # m.fs.filter.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(1)
    # m.fs.filter.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(0.035)
    # m.fs.filter.inlet.flow_mass_phase_comp[0, 'Liq', 'TSS'].fix(120e-6)

    # specify the recovery or removal
    m.fs.filter.removal_fraction_mass_phase_comp["Liq", "TSS"].fix(0.9)
    m.fs.filter.recovery_mass_phase_comp["Liq", "H2O"].fix(0.97)
    m.fs.filter.recovery_mass_phase_comp["Liq", "NaCl"].fix(0.97)

    # Currently the outlet pressure of the waste is unused (i.e. not used in any constraint) so it isn't counted in the
    # degrees of freedom, but if we connected the waste to another unit model then the pressure would be used.
    # So really the unit model has 9 DOF and I'm fixing the last one here.
    m.fs.filter.properties_waste[0].pressure.fix(101325)
    print("DOF after specifying:", degrees_of_freedom(m.fs))

    # the user should provide the scale for the flow rate, so that our tools can ensure the model is well scaled
    # generally scaling factors should be such that if it is multiplied by the variable it will range between 0.01 and 100
    m.fs.properties.set_default_scaling("flow_mass_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e2, index=("Liq", "NaCl")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e4, index=("Liq", "TSS")
    )
    iscale.calculate_scaling_factors(m.fs)  # this utility scales the model

    # solving
    assert_units_consistent(m)  # check that units are consistent
    assert (
        degrees_of_freedom(m) == 0
    )  # check that the degrees of freedom are what we expect

    solver = get_solver()
    results = solver.solve(m, tee=False)
    assert_optimal_termination(results)

    print("second display")
    m.fs.filter.display()

    return m


if __name__ == "__main__":
    main()



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import Suffix, check_optimal_termination
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import InitializationMixin

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Condenser")
class CompressorData(InitializationMixin, UnitModelBlockData):
    """
    Condenser model for MVC
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    def build(self):
        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Add control volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        # complete condensation mass balance
        @self.control_volume.Constraint(
            self.flowsheet().time,
            self.config.property_package.component_list,
            doc="Mass balance",
        )
        def mass_balance(b, t, j):
            lb = b.properties_out[t].get_material_flow_terms("Vap", j).lb
            b.properties_out[t].get_material_flow_terms("Vap", j).fix(lb)
            return b.properties_in[t].get_material_flow_terms(
                "Vap", j
            ) + b.properties_in[t].get_material_flow_terms(
                "Liq", j
            ) == b.properties_out[
                t
            ].get_material_flow_terms(
                "Liq", j
            )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_heat_transfer=True
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type
        )

        # # Add constraints
        @self.Constraint(self.flowsheet().time, doc="Saturation pressure constraint")
        def eq_condenser_pressure_sat(b, t):
            return (
                b.control_volume.properties_out[t].pressure
                >= b.control_volume.properties_out[t].pressure_sat
            )

        # Add ports
        self.add_port(name="inlet", block=self.control_volume.properties_in)
        self.add_port(name="outlet", block=self.control_volume.properties_out)

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        hold_state=False,
        heat=None,
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)
            hold_state : boolean indicating if the inlet conditions should stay fixed
            heat : inital guess for heat transfer out of condenser (negative)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize control volume
        flags = self.control_volume.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # # ---------------------------------------------------------------------
        # check if guess is needed for the heat based on degrees of freedom
        has_guessed_heat = False
        if degrees_of_freedom(self) > 1:
            raise RuntimeError(
                "The model has {} degrees of freedom rather than 0 or 1 (with a guessed heat) for initialization."
                " This error suggests that an outlet condition has not been fixed"
                " for initialization.".format(degrees_of_freedom(self))
            )
        elif degrees_of_freedom(self) == 1:
            if heat is not None:
                self.control_volume.heat.fix(heat)
                has_guessed_heat = True
            else:
                raise RuntimeError(
                    "The model has 1 degree of freedom rather than 0 for initialization."
                    " A common error for this model is not providing a guess for heat in the initialization."
                )

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        if hold_state:
            pass
        else:
            # Release Inlet state
            self.control_volume.release_state(flags, outlvl=outlvl)
            init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))
        if has_guessed_heat:
            self.control_volume.heat.unfix()

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

        return flags

    def _get_performance_contents(self, time_point=0):
        var_dict = {"Heat duty": self.control_volume.heat[time_point]}
        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for (t, j), c in self.control_volume.mass_balance.items():
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[t].flow_mass_phase_comp["Vap", j]
            )
            iscale.constraint_scaling_transform(c, sf)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    NonNegativeReals,
    NegativeReals,
    value,
)

from idaes.core import declare_process_block_class, FlowDirection, useDefault
from idaes.core.util import scaling as iscale
from watertap.core import (
    MembraneChannel1DBlock,
    PressureChangeType,
)
from watertap.core.membrane_channel1d import CONFIG_Template
from watertap.unit_models.osmotically_assisted_reverse_osmosis_base import (
    OsmoticallyAssistedReverseOsmosisBaseData,
    _add_has_full_reporting,
)
import idaes.logger as idaeslog


__author__ = "Adam Atia, Chenyu Wang"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("OsmoticallyAssistedReverseOsmosis1D")
class OsmoticallyAssistedReverseOsmosis1DData(
    OsmoticallyAssistedReverseOsmosisBaseData
):
    """
    Standard 1D OARO Unit Model Class:
    - one dimensional model
    - steady state only
    - single liquid phase only
    """

    CONFIG = CONFIG_Template()

    _add_has_full_reporting(CONFIG)

    def _process_config(self):
        super()._process_config()
        if self.config.transformation_method is useDefault:
            _log.warning(
                "Discretization method was "
                "not specified for the "
                "reverse osmosis module. "
                "Defaulting to finite "
                "difference method."
            )
            self.config.transformation_method = "dae.finite_difference"

        if self.config.transformation_scheme is useDefault:
            _log.warning(
                "Discretization scheme was "
                "not specified for the "
                "reverse osmosis module."
                "Defaulting to backward finite "
                "difference."
            )
            self.config.transformation_scheme = "BACKWARD"

    def _add_membrane_channels_and_geometry(self):

        # Build membrane channel control volume
        channel_kwargs = {
            "dynamic": self.config.dynamic,
            "has_holdup": self.config.has_holdup,
            "area_definition": self.config.area_definition,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args,
            "transformation_method": self.config.transformation_method,
            "transformation_scheme": self.config.transformation_scheme,
            "finite_elements": self.config.finite_elements,
            "collocation_points": self.config.collocation_points,
        }

        self.feed_side = MembraneChannel1DBlock(**channel_kwargs)
        self.permeate_side = MembraneChannel1DBlock(**channel_kwargs)

        self._add_length_and_width()
        add_geometry_kwargs = {
            "length_var": self.length,
            "width_var": self.width,
        }
        self.feed_side.add_geometry(
            flow_direction=FlowDirection.forward, **add_geometry_kwargs
        )
        self.permeate_side.add_geometry(
            flow_direction=FlowDirection.backward, **add_geometry_kwargs
        )
        self._add_area(include_constraint=True)

    def _add_deltaP(self, side):
        if not isinstance(side, str):
            raise TypeError(
                f"{side} is not a string. Please provide a string for the side argument."
            )

        units_meta = self.config.property_package.get_metadata().get_derived_units
        mem_side = getattr(self, side)
        setattr(
            mem_side,
            "deltaP_stage",
            Var(
                self.flowsheet().config.time,
                initialize=-1e5,
                bounds=(-1e6, 0),
                domain=NegativeReals,
                units=units_meta("pressure"),
                doc=f"Pressure drop across {side}",
            ),
        )
        if self.config.pressure_change_type == PressureChangeType.fixed_per_stage:

            @mem_side.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Fixed pressure drop across unit",
            )
            def eq_pressure_drop(b, t, x):
                return mem_side.deltaP_stage[t] == b.length * mem_side.dP_dx[t, x]

        else:

            @mem_side.Constraint(
                self.flowsheet().config.time, doc="Pressure drop across unit"
            )
            def eq_pressure_drop(b, t):
                return mem_side.deltaP_stage[t] == sum(
                    mem_side.dP_dx[t, x] * b.length / b.nfe
                    for x in b.difference_elements
                )

    def _add_mass_transfer(self):

        units_meta = self.config.property_package.get_metadata().get_derived_units

        # mass transfer
        def mass_transfer_phase_comp_initialize(b, t, x, p, j):
            return value(
                self.feed_side.properties[t, x].get_material_flow_terms("Liq", j)
                * self.recovery_mass_phase_comp[t, "Liq", j]
            )

        self.mass_transfer_phase_comp = Var(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=mass_transfer_phase_comp_initialize,
            bounds=(0.0, 1e6),
            domain=NonNegativeReals,
            units=units_meta("mass")
            * units_meta("time") ** -1
            * units_meta("length") ** -1,
            doc="Mass transfer to permeate",
        )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer term",
        )
        def eq_mass_transfer_term(self, t, x, p, j):
            return (
                self.mass_transfer_phase_comp[t, x, p, j]
                == -self.feed_side.mass_transfer_term[t, x, p, j]
            )

        # Feed and permeate-side connection
        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass transfer from feed to permeate",
        )
        def eq_connect_mass_transfer(b, t, x, p, j):
            return (
                b.permeate_side.mass_transfer_term[t, x, p, j]
                == -b.feed_side.mass_transfer_term[t, x, p, j]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            self.difference_elements,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Permeate production",
        )
        def eq_permeate_production(b, t, x, p, j):
            return (
                -b.feed_side.mass_transfer_term[t, x, p, j]
                == b.width * b.flux_mass_phase_comp[t, x, p, j]
            )

    def calculate_scaling_factors(self):

        super().calculate_scaling_factors()

        for (t, x, p, j), v in self.mass_transfer_phase_comp.items():
            sf = iscale.get_scaling_factor(
                self.feed_side.properties[t, x].get_material_flow_terms(p, j)
            )
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)
            v = self.feed_side.mass_transfer_term[t, x, p, j]
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)
            v = self.permeate_side.mass_transfer_term[t, x, p, j]
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)

        if hasattr(self.feed_side, "deltaP_stage"):
            for v in self.feed_side.deltaP_stage.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)
        if hasattr(self.permeate_side, "deltaP_stage"):
            for v in self.permeate_side.deltaP_stage.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Thermophysical property package to be used in conjunction with ASM2d reactions.

Reference:
[1] Henze, M., Gujer, W., Mino, T., Matsuo, T., Wentzel, M.C., Marais, G.v.R.,
Van Loosdrecht, M.C.M., "Activated Sludge Model No.2D, ASM2D", 1999,
Wat. Sci. Tech. Vol. 39, No. 1, pp. 165-182

"""

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    PhysicalParameterBlock,
    StateBlockData,
    StateBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    LiquidPhase,
    Component,
    Solute,
    Solvent,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
import idaes.logger as idaeslog

# Some more information about this module
__author__ = "Andrew Lee, Adam Atia, Xinhong Liu"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ASM2dParameterBlock")
class ASM2dParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._state_block_class = ASM2dStateBlock

        # Add Phase objects
        self.Liq = LiquidPhase()

        # Add Component objects
        self.H2O = Solvent()

        # Soluble species
        self.S_A = Solute(
            doc="Fermentation products, considered to be acetate. [kg COD/m^3]"
        )
        self.S_F = Solute(
            doc="Fermentable, readily bio-degradable organic substrates. [kg COD/m^3]"
        )
        self.S_I = Solute(doc="Inert soluble organic material. [kg COD/m^3]")
        self.S_N2 = Solute(
            doc="Dinitrogen, N2. SN2 is assumed to be the only nitrogenous product of denitrification. [kg N/m^3]"
        )
        self.S_NH4 = Solute(doc="Ammonium plus ammonia nitrogen. [kg N/m^3]")
        self.S_NO3 = Solute(
            doc="Nitrate plus nitrite nitrogen (N03' + N02' -N). SN03 is assumed to include nitrate as well as nitrite nitrogen. [kg N/m^3]"
        )
        self.S_O2 = Solute(doc="Dissolved oxygen. [kg O2/m^3]")
        self.S_PO4 = Solute(
            doc="Inorganic soluble phosphorus, primarily ortho-phosphates. [kg P/m^3]"
        )
        self.S_ALK = Component(doc="Alkalinity, [mol HCO3- per m^3]")

        # Particulate species
        self.X_AUT = Solute(doc="Autotrophic nitrifying organisms. [kg COD/m^3]")
        self.X_H = Solute(doc="Heterotrophic organisms. [kg COD/m^3]")
        self.X_I = Solute(doc="Inert particulate organic material. [kg COD/m^3]")
        self.X_MeOH = Solute(doc="Metal-hydroxides. [kg TSS/m^3]")
        self.X_MeP = Solute(doc="Metal-phosphate, MeP04. [kg TSS/m^3]")
        self.X_PAO = Solute(doc="Phosphate-accumulating organisms. [kg COD/m^3]")
        self.X_PHA = Solute(
            doc="A cell internal storage product of phosphorus-accumulating organisms, primarily comprising poly-hydroxy-alkanoates (PHA). [kg COD/m^3]"
        )
        self.X_PP = Solute(doc="Poly-phosphate. [kg P/m^3]")
        self.X_S = Solute(doc="Slowly biodegradable substrates. [kg COD/m^3]")
        self.X_TSS = Solute(doc="Total suspended solids, TSS. [kg TSS/m^3]")

        # Create sets for use across ASM models and associated unit models
        self.non_particulate_component_set = pyo.Set(
            initialize=[
                "S_A",
                "S_F",
                "S_I",
                "S_N2",
                "S_NH4",
                "S_NO3",
                "S_O2",
                "S_PO4",
                "S_ALK",
                "H2O",
            ]
        )
        self.particulate_component_set = pyo.Set(
            initialize=[
                "X_AUT",
                "X_H",
                "X_I",
                "X_MeOH",
                "X_MeP",
                "X_PAO",
                "X_PHA",
                "X_PP",
                "X_S",
                "X_TSS",
            ]
        )
        self.tss_component_set = pyo.Set(initialize=["X_TSS"])

        # Heat capacity of water
        self.cp_mass = pyo.Param(
            initialize=4182,
            doc="Specific heat capacity of water",
            units=pyo.units.J / pyo.units.kg / pyo.units.K,
        )
        # Density of water
        self.dens_mass = pyo.Param(
            initialize=997,
            doc="Density of water",
            units=pyo.units.kg / pyo.units.m**3,
        )

        # Thermodynamic reference state
        self.pressure_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=101325.0,
            doc="Reference pressure",
            units=pyo.units.Pa,
        )
        self.temperature_ref = pyo.Param(
            within=pyo.PositiveReals,
            mutable=True,
            default=298.15,
            doc="Reference temperature",
            units=pyo.units.K,
        )

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "flow_vol": {"method": None},
                "pressure": {"method": None},
                "temperature": {"method": None},
                "conc_mass_comp": {"method": None},
            }
        )
        obj.define_custom_properties(
            {
                "alkalinity": {"method": None},
            }
        )
        obj.add_default_units(
            {
                "time": pyo.units.s,
                "length": pyo.units.m,
                "mass": pyo.units.kg,
                "amount": pyo.units.kmol,
                "temperature": pyo.units.K,
            }
        )


class _ASM2dStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(
        self,
        state_args=None,
        state_vars_fixed=False,
        hold_state=False,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for property package.

        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:
            flow_vol : value at which to initialize total volumetric flow (default=None)
            alkalinity: value of alkalinity expressed as molar concentration
            conc_mass_comp : value at which to initialize component concentrations (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature (default=None)
            outlvl : sets output level of initialization routine
            state_vars_fixed: Flag to denote if state vars have already been fixed.
                              True - states have already been fixed and
                              initialization does not need to worry
                              about fixing and unfixing variables.
                              False - states have not been fixed. The state
                              block will deal with fixing/unfixing.
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         True - states variables are not unfixed, and
                         a dict of returned containing flags for
                         which states were fixed during initialization.
                         False - state variables are unfixed after
                         initialization by calling the release_state method.

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(self, state_args)

        else:
            # Check when the state vars are fixed already result in dof 0
            for k in self.keys():
                if degrees_of_freedom(self[k]) != 0:
                    raise Exception(
                        "State vars fixed but degrees of freedom "
                        "for state block is not zero during "
                        "initialization."
                    )

        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)

        init_log.info("Initialization Complete.")

    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        """
        Method to relase state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(self, flags)
        init_log.info("State Released.")


@declare_process_block_class("ASM2dStateBlock", block_class=_ASM2dStateBlock)
class ASM2dStateBlockData(StateBlockData):
    """
    StateBlock for calculating thermophysical proeprties associated with the ASM2d
    reaction system.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super().build()

        # Create state variables
        self.flow_vol = pyo.Var(
            initialize=1.0,
            domain=pyo.NonNegativeReals,
            doc="Total volumentric flowrate",
            units=pyo.units.m**3 / pyo.units.s,
        )
        self.pressure = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=101325.0,
            bounds=(1e3, 1e6),
            doc="Pressure",
            units=pyo.units.Pa,
        )
        self.temperature = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=298.15,
            bounds=(273.15, 323.15),
            doc="Temperature",
            units=pyo.units.K,
        )
        self.conc_mass_comp = pyo.Var(
            self.params.solute_set,
            domain=pyo.NonNegativeReals,
            initialize=0.1,
            doc="Component mass concentrations",
            units=pyo.units.kg / pyo.units.m**3,
        )
        self.alkalinity = pyo.Var(
            domain=pyo.NonNegativeReals,
            initialize=1,
            doc="Alkalinity in molar concentration",
            units=pyo.units.kmol / pyo.units.m**3,
        )

    def get_material_flow_terms(self, p, j):
        if j == "H2O":
            return self.flow_vol * self.params.dens_mass
        elif j == "S_ALK":
            # Convert moles of alkalinity to mass assuming all is HCO3-
            return (
                self.flow_vol * self.alkalinity * (61 * pyo.units.kg / pyo.units.kmol)
            )
        else:
            return self.flow_vol * self.conc_mass_comp[j]

    def get_enthalpy_flow_terms(self, p):
        return (
            self.flow_vol
            * self.params.dens_mass
            * self.params.cp_mass
            * (self.temperature - self.params.temperature_ref)
        )

    def get_material_density_terms(self, p, j):
        if j == "H2O":
            return self.params.dens_mass
        elif j == "S_ALK":
            # Convert moles of alkalinity to mass assuming all is HCO3-
            return self.alkalinity * (61 * pyo.units.kg / pyo.units.kmol)
        else:
            return self.conc_mass_comp[j]

    def get_energy_density_terms(self, p):
        return (
            self.params.dens_mass
            * self.params.cp_mass
            * (self.temperature - self.params.temperature_ref)
        )

    def default_material_balance_type(self):
        return MaterialBalanceType.componentPhase

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        return {
            "flow_vol": self.flow_vol,
            "alkalinity": self.alkalinity,
            "conc_mass_comp": self.conc_mass_comp,
            "temperature": self.temperature,
            "pressure": self.pressure,
        }

    def define_display_vars(self):
        return {
            "Volumetric Flowrate": self.flow_vol,
            "Molar Alkalinity": self.alkalinity,
            "Mass Concentration": self.conc_mass_comp,
            "Temperature": self.temperature,
            "Pressure": self.pressure,
        }

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################


import h5py
import numpy as np
import time
import os
import datetime
from datetime import datetime

import logging

__author__ = "Alexander V. Dudchenko (SLAC)"


_log = logging.getLogger(__name__)


def merge_data_into_file(
    file_name,
    backup_file_name,
    directory,
    expected_solved_values=None,
    min_solve_values=None,
    force_rerun=None,
):
    """
    This function checks if there is a sim file, and a back up file.
    if there is a sim file, it backs it up, and checks if full solution set already exsts and copies it over,
    other wise it creates a new group in actual file, and adds new solved data set to it
    """
    run_sweep = True
    if os.path.isfile(file_name) == False:
        create_h5_file(file_name)
    h5file = h5py.File(file_name, "a")

    # check if there is a back up
    if isinstance(backup_file_name, str):
        f_old_solutions = h5py.File(backup_file_name, "r")
        if (
            directory in f_old_solutions
            and "solve_successful" in f_old_solutions[directory]
        ):
            solved_values = sum(
                np.array(
                    f_old_solutions[directory]["solve_successful"]["solve_successful"][
                        ()
                    ]
                )
            )
        else:
            solved_values = None
    else:
        solved_values = None
    if force_rerun:
        _log.info("Forcing a rerun")
        run_sweep = False
    elif force_rerun == False:
        _log.info("Forced to not rerun")
        run_sweep = False
    elif force_rerun == None and solved_values is not None:
        if min_solve_values is not None:
            if min_solve_values <= solved_values:
                run_sweep = False
        elif expected_solved_values == solved_values:
            run_sweep = False
        _log.info(
            "Found {} solved values, expected {} solved values , min {} solved values, re-running == {}".format(
                solved_values, expected_solved_values, min_solve_values, run_sweep
            )
        )
    if run_sweep:
        if directory not in h5file:
            h5file.create_group(directory)
    elif backup_file_name is not None and os.path.isfile(backup_file_name):
        if directory not in h5file:
            h5file.copy(f_old_solutions[directory], directory)
        else:
            _log.warning(
                "Solution already {} exist in file, not copying over back up data".format(
                    directory
                )
            )
        f_old_solutions.close()
    h5file.close()
    return run_sweep


def create_backup_file(file_name, backup_name, h5_dir):
    """used to created file and back up file
    file_name - orignal h5 file
    backup_name - backup name for h5 file if exists
    h5_dir - directory to check in h5 file, if not there creates fresh file, otherwise
    renames existing file"""

    if backup_name is None and os.path.isfile(file_name):
        h5file = h5py.File(file_name, "r")

        if h5_dir in h5file:
            # need to close file before renaming it
            h5file.close()
            date = datetime.now().strftime("%d_%m-%H_%M_%S")
            backup_name = file_name + "_{}_{}".format(date, ".bak")
            if os.path.isfile(backup_name) == False:
                os.rename(file_name, backup_name)
        else:
            # close it in case we did not rename
            h5file.close()
    return backup_name


def create_h5_file(file_name):
    """used to created h5file, retry in case disk is busy"""
    if os.path.isfile(file_name) == False:
        file_created = False
        for i in range(60):
            try:
                f = h5py.File(file_name, "w")
                f.close()
                file_created = True
                break
            except:
                _log.warning("Could note create h5 file {}".format(file_name))
                time.sleep(0.01)  # Waiting to see if file is free to create again
        if file_created == False:
            raise OSError("Could not create file {}".format(file_name))



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a gas-sparged membrane unit.
"""
from types import MethodType
from idaes.core import declare_process_block_class

from watertap.core import pump_electricity, ZeroOrderBaseData
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from pyomo.environ import NonNegativeReals, Var, units as pyunits, Reference
from watertap.core.zero_order_sido import initialize_sido

# Some more information about this module
__author__ = "Adam Atia"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("GasSpargedMembraneZO")
class GasSpargedMembraneZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a gas-sparged membrane.
    This unit is similar to a SIDO, but there is technically a third outlet for gas extraction.
    Three StateBlocks are added with corresponding Ports:

        * properties_inlet
        * properties_treated
        * properties_byproduct

    Two additional variables are added:

        * recovery_vol (indexed by time)
        * removal_frac_mass_comp (indexed by time and component)

    Four additional constraints are added to represent the material balances, with modifications
    to account for gas extraction.

        * water_recovery_equation (indexed by time)
        * flow_balance (indexed by time)
        * solute_removal_equation (indexed by time and solute)
        * solute_treated_equation (indexed by time and solute)

    The build method also sets private attributes on the unit model with references
    to the appropriate initialization and scaling methods to use and to return
    the inlet volumetric flow rate.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "gas_sparged_membrane"

        self._has_recovery_removal = True
        self._initialize = MethodType(initialize_sido, self)
        self._scaling = MethodType(calculate_scaling_factors_gas_extraction, self)

        # Create state blocks for inlet and outlets
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.properties_in = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties at inlet", **tmp_dict
        )

        tmp_dict_2 = dict(**tmp_dict)
        tmp_dict_2["defined_state"] = False

        self.properties_treated = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties of treated water",
            **tmp_dict_2
        )
        self.properties_byproduct = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties of byproduct stream",
            **tmp_dict_2
        )

        # Create Ports
        self.add_port("inlet", self.properties_in, doc="Inlet port")
        self.add_port(
            "treated", self.properties_treated, doc="Treated water outlet port"
        )
        self.add_port(
            "byproduct", self.properties_byproduct, doc="Byproduct outlet port"
        )

        # Add performance variables
        self.recovery_frac_mass_H2O = Var(
            self.flowsheet().time,
            initialize=0.8,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            bounds=(0.0, 1.0000001),
            doc="Mass recovery fraction of water in the treated stream",
        )
        self.removal_frac_mass_comp = Var(
            self.flowsheet().time,
            self.config.property_package.solute_set,
            domain=NonNegativeReals,
            initialize=0.01,
            units=pyunits.dimensionless,
            doc="Solute removal fraction on a mass basis",
        )
        self.gas_mass_influent_ratio = Var(
            self.flowsheet().time,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            doc="Mass flow of gas extracted per mass flow of influent",
        )
        self.flow_mass_gas_extraction = Var(
            self.flowsheet().time,
            domain=NonNegativeReals,
            units=pyunits.kg / pyunits.s,
            doc="Mass flow of hydrogen extracted",
        )
        self._fixed_perf_vars.append(self.gas_mass_influent_ratio)
        self._perf_var_dict[
            "Mass of gas extracted per mass flow of influent(kg/d/(kg/d)"
        ] = self.gas_mass_influent_ratio
        self._perf_var_dict[
            "Mass flow of gas extracted (kg/s))"
        ] = self.flow_mass_gas_extraction

        # Add performance constraints
        # Water recovery
        @self.Constraint(self.flowsheet().time, doc="Water recovery equation")
        def water_recovery_equation(b, t):
            return (
                b.recovery_frac_mass_H2O[t] * b.properties_in[t].flow_mass_comp["H2O"]
                == b.properties_treated[t].flow_mass_comp["H2O"]
            )

        # Flow balance
        @self.Constraint(self.flowsheet().time, doc="Overall flow balance")
        def mass_balance(b, t):
            return (
                sum(
                    b.properties_in[t].flow_mass_comp[j]
                    for j in self.config.property_package.component_list
                )
                == sum(
                    b.properties_treated[t].flow_mass_comp[j]
                    for j in self.config.property_package.component_list
                )
                + sum(
                    b.properties_byproduct[t].flow_mass_comp[j]
                    for j in self.config.property_package.component_list
                )
                + b.flow_mass_gas_extraction[t]
            )

        # Gas extraction
        @self.Constraint(self.flowsheet().time, doc="Gas extraction equation")
        def mass_gas_extraction_equation(b, t):
            return b.flow_mass_gas_extraction[t] == b.gas_mass_influent_ratio[t] * sum(
                b.properties_in[t].flow_mass_comp[j]
                for j in self.config.property_package.component_list
            )

        # Solute removal
        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.solute_set,
            doc="Solute removal equations",
        )
        def solute_removal_equation(b, t, j):
            return (
                b.removal_frac_mass_comp[t, j] * b.properties_in[t].flow_mass_comp[j]
                == b.properties_byproduct[t].flow_mass_comp[j]
            )

        # Solute concentration of treated stream
        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.solute_set,
            doc="Constraint for solute concentration in treated " "stream.",
        )
        def solute_treated_equation(b, t, j):
            return (1 - b.removal_frac_mass_comp[t, j]) * b.properties_in[
                t
            ].flow_mass_comp[j] == b.properties_treated[t].flow_mass_comp[j]

        self._stream_table_dict = {
            "Inlet": self.inlet,
            "Treated": self.treated,
            "Byproduct": self.byproduct,
        }

        self._perf_var_dict["Water Recovery"] = self.recovery_frac_mass_H2O
        self._perf_var_dict["Solute Removal"] = self.removal_frac_mass_comp

        self._get_Q = MethodType(_get_Q_gas_extraction, self)

        self._Q = Reference(self.properties_in[:].flow_vol)
        pump_electricity(self, self._Q)

    # no costing method
    @property
    def default_costing_method(self):
        return lambda *args, **kwargs: None


def calculate_scaling_factors_gas_extraction(self):
    # Get default scale factors and do calculations from base classes
    for t, v in self.water_recovery_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp["H2O"],
                default=1,
                warning=True,
                hint=" for water recovery",
            ),
        )

    for t, v in self.mass_balance.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp["H2O"], default=1, warning=False
            ),
        )  # would just be a duplicate of above

    for (t, j), v in self.solute_removal_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp[j],
                default=1,
                warning=True,
                hint=" for solute removal",
            ),
        )

    for (t, j), v in self.solute_treated_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp[j], default=1, warning=False
            ),
        )  # would just be a duplicate of above
    for t, v in self.mass_gas_extraction_equation.items():
        iscale.constraint_scaling_transform(v, 1e3)


def _get_Q_gas_extraction(self, t):
    return self.properties_in[t].flow_vol



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
High-level tests for the Electrolyte Database (EDB)
"""
import json
import os
import pytest

from pyomo.common.dependencies import attempt_import

from watertap.edb import commands
from watertap.edb.db_api import ElectrolyteDB
from watertap.edb.validate import validate

mongomock, mongomock_available = attempt_import("mongomock")


class MockDB(ElectrolyteDB):
    def __init__(self, db="foo", **kwargs):
        if not mongomock_available:
            pytest.skip(reason="mongomock (EDB optional dependency) not available")

        self._client = mongomock.MongoClient()
        self._db = getattr(self._client, db)
        # note: don't call superclass!
        self._database_name = db
        self._server_url = "mock"


@pytest.fixture
def mockdb():
    return MockDB()


def test_connect(mockdb):
    assert mockdb is not None


def test_load_bootstrap_no_validate(mockdb):
    commands._load_bootstrap(mockdb, do_validate=False)


@pytest.mark.unit
def test_load_bootstrap_data():
    for t in "component", "reaction":
        filename = t + ".json"
        path = commands.get_edb_data(filename)
        input_data = json.load(path.open("r", encoding="utf8"))
        for record in input_data:
            if t == "component":
                record = ElectrolyteDB._process_component(record)
            elif t == "reaction":
                record = ElectrolyteDB._process_reaction(record)
            validate(record, obj_type=t)


@pytest.mark.unit
def test_cloudatlas():
    # this env var should be an encrypted secret in the repository
    passwd = os.environ.get("EDB_CLOUD_PASSWORD", "")
    if not passwd:
        pytest.skip("No password found for MongoDB cloud database")
    url_template = (
        "mongodb+srv://nawi:{passwd}@cluster0.utpac.mongodb.net/test?"
        "authSource=admin&replicaSet=atlas-nxwe6d-shard-0&"
        "readPreference=primary&ssl=true"
    )
    print(f"Connecting to MongoDB cloud server at url={url_template}")
    client = ElectrolyteDB(url=url_template.format(passwd=passwd))



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for ADM1 thermo property package.
Authors: Adam Atia, Alejandro Garciadiego, Xinhong Liu
"""

import pytest

from pyomo.environ import ConcreteModel, Param, value, Var
from pyomo.util.check_units import assert_units_consistent

from idaes.core import MaterialBalanceType, EnergyBalanceType, MaterialFlowBasis

from watertap.property_models.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
    ADM1StateBlock,
)
from idaes.core.util.model_statistics import (
    fixed_variables_set,
    activated_constraints_set,
)

from idaes.core.solvers import get_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = ADM1ParameterBlock()

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.params.state_block_class is ADM1StateBlock

        assert len(model.params.phase_list) == 1
        for i in model.params.phase_list:
            assert i == "Liq"

        assert len(model.params.component_list) == 27
        for i in model.params.component_list:
            assert i in [
                "H2O",
                "S_su",
                "S_aa",
                "S_fa",
                "S_va",
                "S_bu",
                "S_pro",
                "S_ac",
                "S_h2",
                "S_ch4",
                "S_IC",
                "S_IN",
                "S_I",
                "X_c",
                "X_ch",
                "X_pr",
                "X_li",
                "X_su",
                "X_aa",
                "X_fa",
                "X_c4",
                "X_pro",
                "X_ac",
                "X_h2",
                "X_I",
                "S_cat",
                "S_an",
            ]

        assert isinstance(model.params.cp_mass, Param)
        assert value(model.params.cp_mass) == 4182

        assert isinstance(model.params.dens_mass, Param)
        assert value(model.params.dens_mass) == 997

        assert isinstance(model.params.pressure_ref, Param)
        assert value(model.params.pressure_ref) == 101325

        assert isinstance(model.params.temperature_ref, Param)
        assert value(model.params.temperature_ref) == 298.15


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = ADM1ParameterBlock()

        model.props = model.params.build_state_block([1])

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.props[1].flow_vol, Var)
        assert value(model.props[1].flow_vol) == 1

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 298.15

        assert isinstance(model.props[1].anions, Var)
        assert value(model.props[1].anions) == 0.02

        assert isinstance(model.props[1].cations, Var)
        assert value(model.props[1].cations) == 0.04

        assert isinstance(model.props[1].conc_mass_comp, Var)

        assert len(model.props[1].conc_mass_comp) == 24
        Comp_dict = {
            "S_su": 0.012,
            "S_aa": 0.0053,
            "S_fa": 0.099,
            "S_va": 0.012,
            "S_bu": 0.013,
            "S_pro": 0.016,
            "S_ac": 0.20,
            "S_h2": 2.3e-7,
            "S_ch4": 0.055,
            "S_IC": 0.15 * 12,
            "S_IN": 0.13 * 14,
            "S_I": 0.33,
            "X_c": 0.31,
            "X_ch": 0.028,
            "X_pr": 0.10,
            "X_li": 0.029,
            "X_su": 0.42,
            "X_aa": 1.18,
            "X_fa": 0.24,
            "X_c4": 0.43,
            "X_pro": 0.14,
            "X_ac": 0.76,
            "X_h2": 0.32,
            "X_I": 25.6,
        }
        for i in model.props[1].conc_mass_comp:
            assert i in [
                "S_su",
                "S_aa",
                "S_fa",
                "S_va",
                "S_bu",
                "S_pro",
                "S_ac",
                "S_h2",
                "S_ch4",
                "S_IC",
                "S_IN",
                "S_I",
                "X_c",
                "X_ch",
                "X_pr",
                "X_li",
                "X_su",
                "X_aa",
                "X_fa",
                "X_c4",
                "X_pro",
                "X_ac",
                "X_h2",
                "X_I",
            ]
            assert value(model.props[1].conc_mass_comp[i]) == Comp_dict[i]

    @pytest.mark.unit
    def test_get_material_flow_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert model.props[1].get_material_flow_terms(p, j) is (
                    model.props[1].material_flow_expression[j]
                )

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, model):
        for p in model.params.phase_list:
            assert model.props[1].get_enthalpy_flow_terms(p) is (
                model.props[1].enthalpy_flow_expression
            )

    @pytest.mark.unit
    def test_get_material_density_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert model.props[1].get_material_density_terms(p, j) is (
                    model.props[1].material_density_expression[j]
                )

    @pytest.mark.unit
    def test_get_energy_density_terms(self, model):
        for p in model.params.phase_list:
            assert model.props[1].get_energy_density_terms(p) is (
                model.props[1].energy_density_expression
            )

    @pytest.mark.unit
    def test_default_material_balance_type(self, model):
        assert (
            model.props[1].default_material_balance_type()
            == MaterialBalanceType.componentPhase
        )

    @pytest.mark.unit
    def test_default_energy_balance_type(self, model):
        assert (
            model.props[1].default_energy_balance_type()
            == EnergyBalanceType.enthalpyTotal
        )

    @pytest.mark.unit
    def test_get_material_flow_basis(self, model):
        assert model.props[1].get_material_flow_basis() == MaterialFlowBasis.mass

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 6
        for i in sv:
            assert i in [
                "flow_vol",
                "pressure",
                "temperature",
                "conc_mass_comp",
                "anions",
                "cations",
            ]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 6
        for i in sv:
            assert i in [
                "flow_vol",
                "pressure",
                "temperature",
                "conc_mass_comp",
                "anions",
                "cations",
            ]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 6
        for i in sv:
            assert i in [
                "Volumetric Flowrate",
                "Molar anions",
                "Molar cations",
                "Mass Concentration",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(hold_state=False)

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.unit
    def check_units(self, model):
        assert_units_consistent(model)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from watertap.property_models.coagulation_prop_pack import (
    CoagulationParameterBlock,
    CoagulationStateBlock,
)
from watertap.property_models.tests.property_test_harness import PropertyAttributeError
from pyomo.environ import (
    ConcreteModel,
    assert_optimal_termination,
    value,
    Set,
    Param,
    SolverStatus,
    TerminationCondition,
)
from idaes.core import (
    FlowsheetBlock,
    MaterialFlowBasis,
    MaterialBalanceType,
    EnergyBalanceType,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

__author__ = "Austin Ladshaw"

solver = get_solver()

# -----------------------------------------------------------------------------
# Start test class
class TestCoagulationPropPack:
    @pytest.fixture(scope="class")
    def coag_obj(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)
        model.fs.properties = CoagulationParameterBlock()

        return model

    @pytest.mark.unit
    def test_build_properties(self, coag_obj):
        model = coag_obj

        # Check to make sure we have the correct components and phases
        assert isinstance(model.fs.properties.component_list, Set)
        for j in model.fs.properties.component_list:
            assert j in ["H2O", "TDS", "TSS", "Sludge"]
        assert isinstance(model.fs.properties.phase_list, Set)
        for p in model.fs.properties.phase_list:
            assert p in ["Liq"]

        # Check for existance of the state block object
        assert model.fs.properties._state_block_class is CoagulationStateBlock

        # Assert that we have the expected set of parameters
        assert isinstance(model.fs.properties.cp, Param)
        assert isinstance(model.fs.properties.ref_dens_liq, Param)
        assert isinstance(model.fs.properties.dens_slope, Param)
        assert isinstance(model.fs.properties.dens_param_A, Param)
        assert isinstance(model.fs.properties.dens_param_B, Param)
        assert isinstance(model.fs.properties.dens_param_C, Param)
        assert isinstance(model.fs.properties.ref_pressure_correction, Param)
        assert isinstance(model.fs.properties.ref_pressure_slope, Param)
        assert isinstance(model.fs.properties.mu_A, Param)
        assert isinstance(model.fs.properties.mu_B, Param)
        assert isinstance(model.fs.properties.mu_C, Param)

    @pytest.mark.unit
    def test_metadata(self, coag_obj):
        model = coag_obj

        # Create the state block and pull in the default metadata
        model.fs.stream = model.fs.properties.build_state_block([0])
        metadata = model.fs.properties.get_metadata().properties

        # check that properties are not built if not demanded
        for v in metadata.list_supported_properties():
            if metadata[v.name].method is not None:
                if model.fs.stream[0].is_property_constructed(v.name):
                    raise PropertyAttributeError(
                        "Property {v_name} is an on-demand property, but was found "
                        "on the stateblock without being demanded".format(v_name=v.name)
                    )

        # check that properties are built if demanded
        for v in metadata.list_supported_properties():
            if metadata[v.name].method is not None:
                if not hasattr(model.fs.stream[0], v.name):
                    raise PropertyAttributeError(
                        "Property {v_name} is an on-demand property, but was not built "
                        "when demanded".format(v_name=v.name)
                    )

        # check the other stateblock functions
        res = model.fs.stream[0].define_state_vars()
        assert res["flow_mass_phase_comp"] == model.fs.stream[0].flow_mass_phase_comp
        assert res["pressure"] == model.fs.stream[0].pressure
        assert res["temperature"] == model.fs.stream[0].temperature
        assert (
            model.fs.stream[0].get_material_flow_terms("Liq", "H2O")
            == model.fs.stream[0].flow_mass_phase_comp["Liq", "H2O"]
        )
        assert (
            model.fs.stream[0].default_material_balance_type()
            == MaterialBalanceType.componentPhase
        )
        assert (
            model.fs.stream[0].default_energy_balance_type()
            == EnergyBalanceType.enthalpyTotal
        )
        assert model.fs.stream[0].get_material_flow_basis() == MaterialFlowBasis.mass
        assert (
            model.fs.stream[0].get_enthalpy_flow_terms("Liq")
            == model.fs.stream[0].enth_flow["Liq"]
        )

        # NOTE: At this point, all property methods have been touched, so
        #       all will now be built

    @pytest.mark.unit
    def test_stats(self, coag_obj):
        model = coag_obj

        # Check to make sure we have the correct DOF and
        #   check to make sure the units are correct
        assert degrees_of_freedom(model) == 6
        assert_units_consistent(model)

        # Fix some variables and verify it closes the system of equations
        model.fs.stream[0].temperature.fix(298)
        model.fs.stream[0].pressure.fix(101325)
        model.fs.stream[0].flow_mass_phase_comp["Liq", "H2O"].fix(1)
        model.fs.stream[0].flow_mass_phase_comp["Liq", "TSS"].fix(0.01)
        model.fs.stream[0].flow_mass_phase_comp["Liq", "TDS"].fix(0.01)
        model.fs.stream[0].flow_mass_phase_comp["Liq", "Sludge"].fix(0.001)
        assert degrees_of_freedom(model) == 0

    @pytest.mark.unit
    def test_scaling(self, coag_obj):
        model = coag_obj

        # Set some scaling factors and look for 'bad' scaling
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1, index=("Liq", "H2O")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TSS")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e2, index=("Liq", "TDS")
        )
        model.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "Sludge")
        )
        iscale.calculate_scaling_factors(model.fs)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(model))
        assert len(unscaled_var_list) == 0

        # check that all constraints have been scaled
        unscaled_constraint_list = list(iscale.unscaled_constraints_generator(model))
        assert len(unscaled_constraint_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.component
    def test_initialization(self, coag_obj):
        model = coag_obj

        # call the 'calculate_state', which will call initialize and return results
        #       pass var_args as the fixed states from 'test_stats'
        args = {
            ("temperature", None): 298,
            ("pressure", None): 101325,
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.01,
            ("flow_mass_phase_comp", ("Liq", "TSS")): 0.01,
            ("flow_mass_phase_comp", ("Liq", "Sludge")): 0.001,
        }
        results = model.fs.stream.calculate_state(var_args=args)
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        # check to make sure DOF does not change
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve(self, coag_obj):
        model = coag_obj

        # first, check to make sure that after initialized, the scaling is still good
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

        # run solver and check for optimal solution
        results = solver.solve(model)
        assert_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, coag_obj):
        model = coag_obj

        assert value(
            model.fs.stream[0].flow_mass_phase_comp["Liq", "H2O"]
        ) == pytest.approx(1, rel=1e-4)
        assert value(
            model.fs.stream[0].flow_mass_phase_comp["Liq", "TSS"]
        ) == pytest.approx(0.01, rel=1e-4)
        assert value(
            model.fs.stream[0].flow_mass_phase_comp["Liq", "TDS"]
        ) == pytest.approx(0.01, rel=1e-4)
        assert value(
            model.fs.stream[0].flow_mass_phase_comp["Liq", "Sludge"]
        ) == pytest.approx(0.001, rel=1e-4)

        assert value(
            model.fs.stream[0].mass_frac_phase_comp["Liq", "H2O"]
        ) == pytest.approx(0.979431, rel=1e-4)
        assert value(
            model.fs.stream[0].mass_frac_phase_comp["Liq", "TSS"]
        ) == pytest.approx(0.009794, rel=1e-4)
        assert value(
            model.fs.stream[0].mass_frac_phase_comp["Liq", "TDS"]
        ) == pytest.approx(0.009794, rel=1e-4)
        assert value(
            model.fs.stream[0].mass_frac_phase_comp["Liq", "Sludge"]
        ) == pytest.approx(0.0009794, rel=1e-4)

        assert value(model.fs.stream[0].dens_mass_phase["Liq"]) == pytest.approx(
            1013.95727, rel=1e-4
        )
        assert value(model.fs.stream[0].visc_d_phase["Liq"]) == pytest.approx(
            0.0008944, rel=1e-4
        )

        assert value(
            model.fs.stream[0].conc_mass_phase_comp["Liq", "H2O"]
        ) == pytest.approx(993.1021, rel=1e-4)
        assert value(
            model.fs.stream[0].conc_mass_phase_comp["Liq", "TSS"]
        ) == pytest.approx(9.931021, rel=1e-4)
        assert value(
            model.fs.stream[0].conc_mass_phase_comp["Liq", "TDS"]
        ) == pytest.approx(9.931021, rel=1e-4)
        assert value(
            model.fs.stream[0].conc_mass_phase_comp["Liq", "Sludge"]
        ) == pytest.approx(0.9931021, rel=1e-4)


class TestCoagulationPropPackFailures:
    @pytest.fixture(scope="class")
    def coag_obj_fail(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)
        model.fs.properties = CoagulationParameterBlock()

        return model

    @pytest.mark.unit
    def test_default_scaling(self, coag_obj_fail):
        model = coag_obj_fail

        model.fs.stream = model.fs.properties.build_state_block([0])

        # call scaling without setting defaults
        iscale.calculate_scaling_factors(model.fs)

        # check that all variables have scaling factors
        unscaled_var_list = list(iscale.unscaled_variables_generator(model))
        assert len(unscaled_var_list) == 0

        # check that all constraints have been scaled
        unscaled_constraint_list = list(iscale.unscaled_constraints_generator(model))
        assert len(unscaled_constraint_list) == 0

        # check if any variables are badly scaled
        badly_scaled_var_values = {
            var.name: val
            for (var, val) in iscale.badly_scaled_var_generator(
                model, large=1e2, small=1e-2
            )
        }
        assert not badly_scaled_var_values

    @pytest.mark.unit
    def test_state_fail(self, coag_obj_fail):
        model = coag_obj_fail

        args = {
            ("temperature", None): 298,
            ("pressure", None): 101325,
            ("flow_mass_phase_comp", ("Liq", "H2O")): 1,
            ("flow_mass_phase_comp", ("Liq", "TDS")): 0.01,
            ("flow_mass_phase_comp", ("Liq", "TSS")): 0.01,
            ("flow_mass_phase_comp", ("Liq", "Sludge")): 0.001,
        }
        with pytest.raises(ValueError):
            results = model.fs.stream.calculate_state(var_args=args)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import numpy as np
from watertap.tools.MPI.dummy_mpi import DummyCOMM


@pytest.mark.integration
def test_dummy_mpi():
    comm = DummyCOMM
    rank = comm.Get_rank()
    size = comm.Get_size()

    n_arr = 3
    dummy_array1 = np.arange(n_arr)
    comm.Bcast(dummy_array1)
    returned_array = comm.bcast(dummy_array1)
    returned_list = comm.allgather(dummy_array1)
    recvbuf = (np.zeros(n_arr), [n_arr])
    comm.Gatherv(sendbuf=dummy_array1, recvbuf=recvbuf, root=0)

    assert comm is DummyCOMM
    assert rank == 0
    assert size == 1
    assert returned_array is dummy_array1
    assert returned_list == [dummy_array1]
    assert (recvbuf[0] == dummy_array1).all
